﻿using System.Diagnostics;

namespace OBSBlur.Window;

public partial class WindowManager : IDisposable
{
    private readonly HashSet<WindowInfo> _windows = new(1024);
    private List<IntPtr> _zOrder = GetWindowZOrder();
    public IntPtr[] WindowZOrder => _zOrder.ToArray();
    public WindowInfo[] WindowInfos => _windows.ToArray();
    public int UpdateInterval = 10;
    private bool _keepUpdating = true;
    public WindowManager()
    {
        Thread t = new (() =>
        {
            while (_keepUpdating)
            {
                WindowInfo[] windowInfoBefore = WindowInfos;
                IntPtr[] zOrderBefore = WindowZOrder;
                _windows.Clear();
                EnumWindows(GetWindowInfo, 0);
                _zOrder = GetWindowZOrder();
                
                WindowsUpdated?.Invoke(WindowInfos);
                WindowInfo[] changedWindows = WindowInfos.Where(info => !windowInfoBefore.Contains(info)).ToArray();
                if(changedWindows.Length > 0)
                    WindowsChanged?.Invoke(windowInfoBefore.Where(b => changedWindows.Select(c => c.WindowHandle).Contains(b.WindowHandle)).ToArray(), changedWindows);
                if (zOrderBefore.Length != WindowZOrder.Length)
                    ZOrderChanged?.Invoke(WindowZOrder);
                else
                {
                    IntPtr[] zOrderAfter = WindowZOrder;
                    for(int i = 0; i < zOrderBefore.Length; i++)
                        if (!zOrderBefore[i].Equals(zOrderAfter[i]))
                        {
                            ZOrderChanged?.Invoke(zOrderAfter);
                            break;
                        }
                            
                }
                
                Thread.Sleep(UpdateInterval);
            }
        });
        t.Start();
    }

    public delegate void WindowsUpdatedHandler(WindowInfo[] windowInfos);
    public event WindowsUpdatedHandler WindowsUpdated;

    public delegate void WindowsChangedHandler(WindowInfo[] before, WindowInfo[] after);
    public event WindowsChangedHandler WindowsChanged;

    public delegate void WindowsZOrderChangedHandler(IntPtr[] newOrder);

    public event WindowsZOrderChangedHandler ZOrderChanged;

    private bool GetWindowInfo(IntPtr windowHandle, IntPtr lParam)
    {
        bool isVisible = IsWindowVisible(windowHandle);
        //Do not add if Window is not a drawable
        if (!isVisible)
            return true;

        WindowPlacement placement = new ();
        GetWindowPlacement(windowHandle, ref placement);
        
        uint pid;
        GetWindowThreadProcessId(windowHandle, out pid);
        Process processInfo = Process.GetProcessById((int)pid);
        //nvcontainer does not have a window name
        if (processInfo.ProcessName.Equals("nvcontainer"))
            return true;
        string windowTitle = GetWindowTextRaw(windowHandle);
        //Do not add, if window does not have a name
        if (windowTitle.Length < 1)
            return true;

        GetWindowRect(windowHandle, out Rectangle rect);

        _windows.Add(new WindowInfo(windowHandle, windowTitle, processInfo, placement.ShowCmd, rect));
        return true;
    }

    public Rectangle GetDesktopRectangle()
    {
        GetWindowRect(GetDesktopWindow(), out Rectangle rect);
        return rect;
    }

    public void Dispose()
    {
        _keepUpdating = false;
    }
}

